package solution;

import java.awt.geom.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import problem.*;
import tester.Tester;

public class ConfigGenerator {
	
	private Tester tester = new Tester();
	private List<Obstacle> objects;
	private int precision = 5000;
	private Random generator = new Random(2015);
	
	public ConfigGenerator(List<Obstacle> worldObjects) {
		objects = new ArrayList<>(worldObjects);
	}
	
	public ConfigGenerator(int precision, List<Obstacle> worldObjects) {
		this.precision = precision;
		objects = new ArrayList<>(worldObjects);
	}
	
	/** return a list of possible configs at the xy location of the
	 * point given
	 * @param baseCfg
	 * @return
	 */
	public List<ArmConfig> getPossibleConfigs(ArmConfig baseCfg) {
		List<ArmConfig> possibilities = new ArrayList<>();
		int i = 0;
		
		if (validConfig(baseCfg)) {
		    possibilities.add(baseCfg);
		}
		
		while (possibilities.size() < 100 && i < precision) {
			List<Double> joints;
			List<Double> newJoints = new ArrayList<>();
			// our range is 300 degrees expressed in radians
			Double range = 300.0 * Math.PI / 180.0;
			ArmConfig possibility = new ArmConfig(baseCfg);
			
			joints = possibility.getJointAngles();
			for (@SuppressWarnings("unused") Double joint : joints) {
				newJoints.add(generator.nextDouble() * range - (range / 2));
			}
			possibility.setJointAngles(newJoints);
			if (validConfig(possibility)) {
			    possibilities.add(possibility);
			}
			i++;
		}
		
		return possibilities;
	}
	
	/** Return a list of possible configurations in a cloud around the base xy given
	 * 
	 */
	public List<ArmConfig> getConfigCloud(ArmConfig baseCfg) {
		List<ArmConfig> possibilities = new ArrayList<>();
		Point2D origin = baseCfg.getBase();
		List<Point2D> cloud;
		List<Double> initJoints = baseCfg.getJointAngles();
		cloud = generateCloud(origin);
		
		for (Point2D base : cloud) {
			possibilities.addAll(getPossibleConfigs(new ArmConfig(base, initJoints)));
		}
		return possibilities;
	}

	private List<Point2D> generateCloud(Point2D origin) {
		List<Point2D> cloud = new ArrayList<>();
		
		cloud.add(origin);
		for (int i = 0; i < 100; i++) {
			double dx;
			double dy;
			dx = 0.05 * Math.random() - 0.025;
			dy = 0.05 * Math.random() - 0.025;
			cloud.add(new Point2D.Double(origin.getX() + dx, origin.getY() + dy));
		}
		return cloud;
	}

	/** Return a list of possible configurations in a line between the two base xy
	 * locations given
	 * 
	 */
	public List<ArmConfig> getPossibleConfigsBetween(ArmConfig currentCfg, ArmConfig goalCfg) {
		List<ArmConfig> possibilities = new ArrayList<>();
		Point2D currentBase = currentCfg.getBase();
		Point2D goalBase = goalCfg.getBase();
		double theta;
		double newX;
		double newY;
		
		if (currentBase.distance(goalBase) < 0.15) {
			possibilities.addAll(getConfigCloud(currentCfg));
			possibilities.addAll(getConfigCloud(goalCfg));
			return possibilities;
		}
		
		possibilities.addAll(getConfigCloud(currentCfg));
		// work out the location of our new base, 0.1 in between the two
		theta = Math.atan2(goalBase.getY() - currentBase.getY(), 
						   goalBase.getX() - currentBase.getX());
		newX = currentBase.getX() + 0.1 * Math.cos(theta);
		newY = currentBase.getY() + 0.1 * Math.sin(theta);
		
		currentBase.setLocation(newX, newY);
		possibilities.addAll(getPossibleConfigsBetween(new ArmConfig(currentBase, currentCfg.getJointAngles()), goalCfg));
		
		return possibilities;
	}
	
	/**
     * Returns true iff the configuration does NOT collide with anything
     * (i.e. itself, world boundary or object)
     */
    public boolean validConfig(ArmConfig cfg) {
		return !tester.hasCollision(cfg, objects) && 
				!tester.hasSelfCollision(cfg) && 
				tester.fitsBounds(cfg);
    }
    
    public List<ArmConfig> generateObstacleUniverse(ProblemSpec ps) {
    	List<Point2D> pointsOfInterest = new ArrayList<>();
    	List<ArmConfig> nodeList = new ArrayList<>();
    	pointsOfInterest = getCorners(ps.getObstacles());

    	for (Point2D base : pointsOfInterest) {
    		nodeList.addAll(getConfigCloud(new ArmConfig(base, ps.getInitialState().getJointAngles())));
    	}
    	return nodeList;
    }
        
    private List<Point2D> getCorners(List<Obstacle> objects) {
    	List<Point2D> pointsOfInterest = new ArrayList<>();
    	for (Obstacle object : objects) {
    		double [] coords;
    		double xMin, xMax, yMin, yMax;
    		Point2D point;
    		coords = object.getCoords();
    		xMax = coords[0] + 0.001;
    		yMax = coords[1] + 0.001;
    		xMin = coords[2] - 0.001;
    		yMin = coords[3] - 0.001;
    		if (xMax > 1) {
    			xMax = 1;
    		}
    		if (yMax > 1) {
    			yMax = 1;
    		}
    		if (xMin < 0) {
    			xMin = 0;
    		}
    		if (yMin < 0) {
    			yMin = 0;
    		}

    		//We grab the corners as immediate POIs
    		point = new Point2D.Double(xMax, yMax);
    		pointsOfInterest.add(point);
    		point = new Point2D.Double(xMax, yMin);
    		pointsOfInterest.add(point);
    		point = new Point2D.Double(xMin, yMax);
    		pointsOfInterest.add(point);
    		point = new Point2D.Double(xMin, yMin);
    		
    	}
    	return pointsOfInterest;
    }
    
	public List<ArmConfig> generateRandomUniverse(ProblemSpec ps) {
		List<ArmConfig> universe = new ArrayList<>();
		double x;
		double y;
		
		while (universe.size() < 20000) {
			x = Math.random();
			y = Math.random();
			universe.addAll(getPossibleConfigs(new ArmConfig(new Point2D.Double(x, y), ps.getInitialState().getJointAngles())));
		}
		universe.add(ps.getInitialState());
		universe.add(ps.getGoalState());
		return universe;		
	}
}
