package solution;

import java.awt.geom.Point2D;
import java.util.*;

import problem.ArmConfig;
import problem.Obstacle;
import tester.Tester;

public class LocalPlanner {
    private final Integer FULL = 0;
	private final Integer ARM_ONLY = 1;
    private final Integer BASE_ONLY = 2;
    private List<Obstacle> objects;
    private Tester tester = new Tester(); 
	
	public LocalPlanner() {
		
	}
	
	public void setObjects(List<Obstacle> list) {
		this.objects = list;
	}
	/**
     * addPrimitiveStep will populate our final solution with all of the steps
     * to go from one configuration to the next
     * 
     * @param robotPos is a stack containing all of the robot's moves until this
     * point
     * @param nextDest is the place we're going to next
     */
    private void addPrimitiveStep(Stack<ArmConfig> robotPos,  
            ArmConfig nextDest, Integer setting) {
        ArmConfig currentCfg = robotPos.peek();
        ArmConfig newCfg = new ArmConfig(currentCfg);
        Point2D goalCfg = nextDest.getBase();
        List<Double> newAngles = newCfg.getJointAngles();
        List<Double> goalAngles = nextDest.getJointAngles();
        
        // if we are allowed to move the base (setting = FULL or BASE_ONLY), we move it
        if (setting != ARM_ONLY) {
	        if (currentCfg.getBase().distance(nextDest.getBase()) 
	                > Tester.MAX_BASE_STEP) {
	            Double theta;
	            Point2D oldBase = currentCfg.getBase();
	            Point2D newBase;
	            // Work out (in rads) the angle between our two points in polars
	            theta = Math.atan2(goalCfg.getY() - newCfg.getBase().getY(), 
	                               goalCfg.getX() - newCfg.getBase().getX());
	            // with the direction we are travelling, we need to go no further
	            // than 0.0099, hopefully Math.sin and Math.cos round down
	            newBase = new Point2D.Double (
	                    oldBase.getX() + Tester.MAX_BASE_STEP * Math.cos(theta), 
	                    oldBase.getY() + Tester.MAX_BASE_STEP * Math.sin(theta));
	            newCfg.setBase(newBase);
	        }
        }
        
        // if we are allowed to move the arms (setting = FULL or ARM_ONLY), we move them. 
        if (setting != BASE_ONLY) {
	        for (int i = 0; i < newAngles.size(); i++) {
	            Double theta;
	            if (newAngles.get(i) < goalAngles.get(i)) {
	                //move angles in radians by 0.1 degrees
	                theta = newAngles.get(i) + Tester.MAX_JOINT_STEP;
	                newAngles.set(i, theta);
	            } else if (newAngles.get(i) > goalAngles.get(i)) {
	                theta = newAngles.get(i) - Tester.MAX_JOINT_STEP;
	                newAngles.set(i, theta);
	            }
	        }
	        newCfg.setJointAngles(newAngles);
        }

        robotPos.push(newCfg);
    }
    
    /**
     * Test to see if there's a direct solution from orig to goal
     * 
     * @param orig
     * @param goal
     * @return true if a solution exists
     */
     private boolean directSolution(ArmConfig orig, ArmConfig goal) {
        if (orig.getBase().distance(goal.getBase()) > Tester.MAX_BASE_STEP) {
            return false;
        }

        for (int i = 0; i < orig.getJointAngles().size(); i++) {
            Double angle1 = orig.getJointAngles().get(i);
            Double angle2 = goal.getJointAngles().get(i);
            if (Math.abs(angle1 - angle2) > Tester.MAX_JOINT_STEP ) {
                return false;
            }
        }
        return true;
    }
     
    /**
     * Returns true iff the configuration does NOT collide with anything
     * (i.e. itself, world boundary or object)
     */
    private boolean validStep(ArmConfig cfg) {
		return !tester.hasCollision(cfg, objects) && 
				!tester.hasSelfCollision(cfg) && 
				tester.fitsBounds(cfg);
    }
	
	/**
	 * Attempts to construct a path using a series of primitive steps
	 * 
	 * @param path is the path we are constructing
	 * @param goalNode is the place we're going to
	 * @return true iff we were successful
	 */
	private boolean moveNaive(Stack<ArmConfig> path, ArmConfig goalNode) {
	    Stack<ArmConfig> attempt = new Stack<>();
      
	    attempt.push(path.peek());
		while (!attempt.peek().equals(goalNode)) {
    		addPrimitiveStep(attempt, goalNode, FULL);
    		if (directSolution(attempt.peek(), goalNode)) {
    			attempt.push(goalNode);
    		}

    		if (!validStep(attempt.peek())) {
    			return false;
    		}
    	}
		path.addAll(attempt);
		return true;
	}
	
	/**
	 * Attempts to construct a path using a series of primitive steps. 
	 * This time we try moving the arm first, then the base
	 * lib/exe/fetch.php/ai:additionaltests-prob.tar.gz
	 * @param path is the path we are constructing
	 * @param goalNode is the place we're going to
	 * @return true iff we were successful
	 */
	private boolean armFirst(Stack<ArmConfig> path, ArmConfig goalNode) {
		ArmConfig armGoal = new ArmConfig(goalNode);
		armGoal.setBase(path.peek().getBase());
		Stack<ArmConfig> attempt = new Stack<>();
		
		attempt.push(path.peek());
		while (!attempt.peek().equals(armGoal)) {
			addPrimitiveStep(attempt, armGoal, ARM_ONLY);
			if (directSolution(attempt.peek(), armGoal)) {
				attempt.push(armGoal);
			}
			if (!validStep(attempt.peek())) {
    			return false;
    		}
		}
		
		if (moveNaive(attempt, goalNode)) {
		    path.addAll(attempt);
		    return true;
		}
		return false;
	}
	
	/**
	 * Attempts to construct a path using a series of primitive steps.
	 * This time we try moving the base first, then the arms
	 * 
	 * @param path is the path we are constructing
	 * @param goalNode is the place we're going to
	 * @return true iff we were successful
	 */
	private boolean baseFirst(Stack<ArmConfig> path, ArmConfig goalNode) {
		ArmConfig posGoal = new ArmConfig(goalNode);
		posGoal.setJointAngles(path.peek().getJointAngles());
		Stack<ArmConfig> attempt = new Stack<>();
		
		attempt.push(path.peek());
		while (!attempt.peek().equals(posGoal)) {
			addPrimitiveStep(attempt, posGoal, BASE_ONLY);
			if (directSolution(attempt.peek(), posGoal)) {
				attempt.push(posGoal);
			}
			if (!validStep(attempt.peek())) {
    			return false;
    		}
		}
		
		if (moveNaive(attempt, goalNode)) {
		    path.addAll(attempt);
		    return true;
		}
		return false;
	}
	
	public boolean getThere(Stack<ArmConfig> path, ArmConfig goalNode) {
		
		if (moveNaive(path, goalNode)) {
    		return true;
    	}
    	if (baseFirst(path, goalNode)) {
    		return true;
    	}
    	if (armFirst(path, goalNode)) {
    		return true;
    	}
		return false;
	}
	
    public int edgeExists(ArmConfig node1, ArmConfig node2) {
    	Stack<ArmConfig> path = new Stack<>();
    	path.push(node1);
    	if (getThere(path, node2)) {
    		return path.size();
    	}
    	return 0;
    }
}
